#ifndef MTQUEUE_H
#define MTQUEUE_H
///////////////////////////////////////////////////////////////
//  mtqueue.h  -- thread-safe template-based queue           //
//  ver 1.1                                                  //
//  Language:     Visual C++, ver 6.0                        //
//  Platform:     Micron Dual Pentium Pro 200, Win NT 4.0    //
//  Application:  CSE691 - Project #3 prototype              //
//  Author:       Jim Fawcett, CST 2-187, Syracuse Univ      //
//                (315) 443-3948, fawcett@syr.edu            //
///////////////////////////////////////////////////////////////
/*
  Class Operation
  ===============
  The mtqueue class accepts messages with an enQ(m) command where
  m is the message to be queued.  It destructively returns the
  message with deQ().  deQ() returns messages in the order they
  were enqueued.  enQ() and deQ() operations are protected with
  a mutex, unique to each class object.  So they are efficiently 
  synchronized under reads and writes to multiple queues.

  mtqueue is implemented with a doubly linked list of nodes.  Each
  node holds pointers to its predecessor and successor, and a
  message value, so the queue is simply a class holding head and 
  tail pointers, used by enQ(...) and deQ().
    
                                       tail
    arrival      +------+      +------+  |   +------+   server
      end        |node 1|      |node 2|  |   |node 3|    end
                 +------+      +------+  +-->+------+
      head ----->| next |----->| next |----->| next |----+
            +----| prev |<-----| prev |<-----| prev |    |
            |    |  t1  |      |  t2  |      |  t3  |   \ /
           \ /   +------+      +------+      +------+

  enQ(...) adds a node to the head of the list, logically equivalent
  to adding an element to the end of the queue.  deQ() removes an
  element from the tail of the list, logically equivalent to removing
  an element from the beginning of the queue.

    mtqueue<string> sq;   // make a queue to hold string messages
    sq.enQ(s);            // enqueues a message m
    string s = sq.deQ();  // dequeues a message
*/
///////////////////////////////////////////////////////////////
//  Design Notes                                             //
///////////////////////////////////////////////////////////////
/*
    If you want to use this queue class in an application
    that will run for a long time, doing lots of enQ(val)s
    and deQ()s, you should prevent the class from continuously
    newing and deleting nodes.  this will cause the heap to get
    fragmented and performance will suffer, possibly failing 
    due to memory allocation failures.

    The approach I intend to use when I get around to it is to
    keep a list of free nodes.  When deQuing, instead of delet-
    ing the node, put it on a free list.  When enQing, check
    to see if the free list has any nodes before newing one.
    This way the queue will allocate nodes until the queue gets
    to its maximum length.  Then as its length varies with time
    it will simply reuse nodes on the free list.
*/
//

///////////////////////////////////////////////////////////////
//  Build Instructions                                       //
///////////////////////////////////////////////////////////////
//  Files Required:                                          //
//    mtqueue.h, mtqueue.cpp                                 //
//  Compiler Command:                                        //
//    cl -GX -DTEST_MTQUEUE mtqueue.cpp                      //
///////////////////////////////////////////////////////////////
/*
    Maintenance History:
    ====================
    ver 1.1 : 30 November 1999
      - changed deQ() to deQ(T &t), e.g., pass out by reference
        rather than copy out.  This was done to improve run
        time performance.
      - added peekQ(T &t) to look at next element without
        removal
      - added test for memory allocation failure in enQ(...)
      - added synchronization with class mutex to destructor
      - added private declarations of copy ctor and assignment
        operator to prevent compiler from generating them
      - added invariant test
    ver 1.0 : 14 November 1998
      - developed this first release by modifying
        queue.h and queue.cpp

    Future Modifications:
    =====================
    - Use a free list of nodes to minimize heap fragmentation.
      See the Design Note on the previous page for a brief
      discussion of this topic.

*/

//

#include <iostream>
#include <windows.h>
#include <process.h>

using namespace std;

template <class T> class node {

  public:
    node(const T &t, node<T> *&link);
    ~node();
    T& value();
    node<T>*& prev();   // read or write prev pointer
    node<T>*& next();   // read or write next pointer

  private:
    node<T> *_prev;
    node<T> *_next;
    T _t;
};

template <class T> 
inline node<T>::~node() { }

template <class T> 
inline T& node<T>::value() { return _t; }

template <class T> 
inline node<T>*& node<T>::prev() { return _prev; }

template <class T> 
inline node<T>*& node<T>::next() { return _next; }


template <class T> class mtqueue {

  public:
    mtqueue();
    ~mtqueue();
    void enQ(const T &t);
    void deQ(T &t);
    void peekQ(T &t);
    int  length();
    bool invariant();

  private:
    node<T> *head;
    node<T> *tail;
    int len;
    HANDLE mutex;
  // prevent compiler generated copies and assignments
    mtqueue(const mtqueue<T> &q);
    mtqueue<T>& operator=(const mtqueue<T> &q);
};

template <class T> inline int mtqueue<T>::length() { return len; }


//

//----< node constructor >-------------------------------------
//
//  construct node, give it a value, and link to
//  head of the list
//
template <class T> node<T>::node(const T &t, node<T> *&link) {

  _t = t;         // store value
  _prev = NULL;   // this is at head of list so there is no prev
  _next = link;   // link to old first node
}

//----< queue constructor >------------------------------------
//
//  construct queue by setting both head and tail to null
//  signifying an empty queue
//
template <class T> 
mtqueue<T>::mtqueue() : head(NULL), tail(NULL), len(0) { 

  mutex = CreateMutex(0,FALSE,0);
}

//----< queue destructor >-------------------------------------
//
//  walk down list and delete each node
//
template <class T> mtqueue<T>::~mtqueue() {

  WaitForSingleObject(mutex,INFINITE);
  while(head != NULL) {
    node<T> *temp = head;
    head = head->next();
    delete temp;
  }
  head = tail = NULL;
  len = 0;
  ReleaseMutex(mutex);
  CloseHandle(mutex);
}

//

//----< add element to back of queue >-------------------------
//
//  create node and link to the head of queue list
//
template <class T> void mtqueue<T>::enQ(const T &t) {

  WaitForSingleObject(mutex,INFINITE);
  head = new node<T>(t,head);     // create node pointing to old head
  if(head==NULL) {
    cout << "\n  --- memory alloc failure in mtqueue::enQ(...) ---\n";
    cout.flush();
    exit(1);
/////////////////////////////////////////////////////////////////////////
//  exceptions are great for handling errors, but slow down processing //
//    throw "\n --- memory alloc failure in mtqueue::enQ(...) ---\n";  //
/////////////////////////////////////////////////////////////////////////
  }
  if(tail != NULL)                // - if there was a node in queue then
    head->next()->prev() = head;  //   make new node its predecessor
  else                            // - otherwise this is the first node
    tail = head;                  //   so make tail point to it
  len++;

  #ifdef DEBUG
  if(!invariant())
    errorOut("\n  --- queue fails invariant test in enQ ---\n",true);
  if(!t.invariant())
    errorOut("\n  --- message fails invariant test in deQ ---\n",true);
  #endif

  ReleaseMutex(mutex);
}

//

//----< remove element from front of queue >-------------------
//
//  retrieve value of last node on list, e.g., first in queue
//  and eliminate node
//
template <class T> void mtqueue<T>::deQ(T &t) {

  WaitForSingleObject(mutex,INFINITE);
  if(tail == NULL) {
    errorOut("\n  --- deQ on empty Q ---\n",true);
    ReleaseMutex(mutex);
    return;
  }
  t = tail->value();              // retrieve last node's value
  if(tail != head) {              // if this is not the only node
    tail = tail->prev();          //   make new tail pointing to its
    delete tail->next();          //   predecessor and delete node,
    tail->next() = NULL;          //   make new last node nextless
  }
  else {                          // otherwise this is only node
    delete tail;                  //   delete node and make list empty
    head = tail = NULL;           //   by nulling its head and tail
  }
  len--;
  
  #ifdef DEBUG
  if(!invariant()) {
    errorOut("\n  --- queue fails invariant test in deQ ---\n",true);
  }  
  if(!t.invariant())
    errorOut("\n  --- message fails invariant test in deQ ---\n",true);
  #endif

  ReleaseMutex(mutex);
  return;
}

//----< peek at first element in queue (last on list) >--------

template <class T> inline
void mtqueue<T>::peekQ(T &t) { 
  
  t = tail->value();

  #ifdef DEBUG
  if(!t.invariant())
    errorOut("\n  --- message fails invariant test in peekQ ---\n",true);
  #endif

}

//

//----< test queue invariant >---------------------------------
//
//  - walk down queue and check end conditions
//  - walk back up queue and check beginning conditions
//
template <class T> 
bool mtqueue<T>::invariant() {

  if(len == 0) {
    if(head != NULL || tail != NULL)
      return false;
  }
  else if(len == 1) {
    if(head != tail || head->prev() != NULL || tail->next() != NULL)
      return false;
  }
  else {
    node<T> *pNode = head;
    int count = 0;
    while(++count < len) {
      pNode = pNode->next();
    }
    if(pNode != tail || pNode->next() != NULL)
      return false;
    while(--count > 0) {
      pNode = pNode->prev();
    }
    if(pNode != head || pNode->prev() != NULL)
      return false;
  }
  return true;
}

#endif
